(declare (usual-integrations))

(define (rewrite/top-level expression)
  (within-rewriter
   (lambda ()
     (make-lambda
      'expression
      '(initial-continuation initial-fuel initial-store) '() #F () ()
      (rewrite/expression expression
			  'initial-fuel
			  'initial-store
			  '()
			  0
			  (named-continuation 'initial-continuation))))))

(define (try p)
  (define (doit e)
    (let ((syn (syntax&integrate e '((usual-integrations)))))
      (pp syn)
      (rewrite/top-level syn)))
  (cond ((pair? p) (doit p))
	((procedure? p) (doit (procedure-lambda p)))
	(else (rewrite/top-level p))))

(define (rewrite/expression expression fuel store bound ticks cont)
  ((scode-walk rewrite-walker expression)
   expression fuel store bound ticks cont))

;; rewrite: expression * fuel * store * bound-vars * ticks * cont -> expression
;; cont: expression * fuel * store * ticks -> expression

(define cost:unassigned? 1)
(define cost:local-assignment 1)
(define cost:global-assignment 2)
(define cost:global-definition 10)
(define cost:local-reference 0)
(define cost:global-reference 2)
(define cost:if 1)
(define cost:lambda 2)
(define (cost:call n) n 1)
(define (cost:primitive p) p 1)

(define cost:max-sequence 10)
(define cost:max-expression 10)

(define rewrite-walker)

(define (initialize-package!)
  (set! rewrite-walker
	(make-scode-walker
	 rewrite/constant
	 `((ACCESS ,rewrite/access)
	   (ASSIGNMENT ,rewrite/assignment)
	   (COMBINATION ,rewrite/combination)
	   (COMMENT ,rewrite/comment)
	   (CONDITIONAL ,rewrite/conditional)
	   (DEFINITION ,rewrite/definition)
	   (DELAY ,rewrite/delay)
	   (DISJUNCTION ,rewrite/disjunction)
	   (IN-PACKAGE ,rewrite/in-package)
	   (LAMBDA ,rewrite/lambda)
	   (SEQUENCE ,rewrite/sequence)
	   (THE-ENVIRONMENT ,rewrite/the-environment)
	   (UNASSIGNED? ,rewrite/unassigned?)
	   (VARIABLE ,rewrite/variable))))
  unspecific)

(define *shared-definitions*)		; These go in load environment
(define *definition-names*)		; The names defined anywhere
(define *cache-map*)

(define (within-rewriter thunk)
  (fluid-let ((*cache-map* (make-eq-hash-table))
	      (*definition-names* '())
	      (*shared-definitions* '()))
    (thunk)))

;;(define (within-rewriter thunk)
;;  (set! *cache-map* (make-eq-hash-table))
;;  (thunk))

(define (named-continuation name) name)
(define (named-continuation? cont) (symbol? cont))

(define (make-fuel-var)  (generate-uninterned-symbol ".fuel."))
(define (make-store-var) (generate-uninterned-symbol ".store."))
(define (make-cont-var)  (generate-uninterned-symbol ".cont."))

(define (fuel-adjustment fuel ticks)
  (if (zero? ticks)
      (make-variable fuel)
      (make-combination fix:- (list (make-variable fuel) ticks))))

(define (->value cont)
  (if (named-continuation? cont)
      (lambda (value fuel store ticks)
	(make-combination
	 (make-variable 'call/r)
	 (list value
	       (fuel-adjustment fuel ticks)
	       (make-variable store)
	       (make-variable cont))))
      cont))

(define (make-compiled-cache name)
  (or (hash-table/get *cache-map* name #F)
      (let ((cache (vector name #F #F)))
	(hash-table/put! *cache-map* name cache)
	cache)))


(define (rewrite/global-reference name fuel store bound ticks cont)
  bound
  ((->value cont)
   (make-combination (make-variable 'global-ref)
		     (list (make-variable store)
			   (make-compiled-cache name)))
   fuel store (+ ticks cost:global-reference)))

(define (rewrite/variable expression fuel store bound ticks cont)
  (let ((name (variable-name expression)))
    (if (memq name bound)
	((->value cont)
	 (make-variable name) fuel store (+ ticks cost:local-reference))
	(rewrite/global-reference name fuel store bound ticks cont))))

(define (rewrite/access expression fuel store bound ticks cont)
  (if (eq? (access-environment expression) system-global-environment)
      (rewrite/global-reference (access-name expression)
				fuel store bound ticks cont)
      (error "Sorry, non-global ACCESS not allowed:" (unsyntax expression))))


(define (rewrite/global-assignment name value fuel store ticks cont)
  (make-sequence 
   (list (make-combination (make-variable 'global-set!)
			   (list (make-variable store)
				 (make-compiled-cache name)
				 value))
	 ((->value cont)
	  unspecific fuel store (+ ticks cost:global-assignment)))))

(define (rewrite/global-assignment name value fuel store ticks cont)
  ((->value cont)
   (make-combination (make-variable 'global-set!)
		     (list (make-variable store)
			   (make-compiled-cache name)
			   value))
   fuel store (+ ticks cost:global-assignment)))

(define (rewrite/global-definition name value fuel store ticks cont)
  ((->value cont)
   (make-combination (make-variable 'global-define!)
		     (list (make-variable store)
			   (make-compiled-cache name)
			   value))
   fuel store (+ ticks cost:global-definition)))


(define (rewrite/assignment expression fuel store bound ticks cont)
  (let ((name  (assignment-name expression)))
    (rewrite/expression
     (assignment-value expression)
     fuel store bound ticks
     (lambda (value fuel store ticks)
       (if (memq name bound)
	   (make-sequence
	    (list (make-assignment name value)
		  ((->value cont) unspecific fuel store (+ ticks cost:local-assignment))))
	   (rewrite/global-assignment name value fuel store ticks cont))))))


(define (rewrite/lambda expression fuel store bound ticks cont)
  (lambda-components expression
    (lambda (name required optional rest auxiliary declarations body)
      (let ((k-var (make-cont-var))
	    (f-var (make-fuel-var))
	    (s-var (make-store-var)))
	(let ((body*
	       (rewrite/body body f-var s-var auxiliary
			     (append required optional (if rest (list rest) ())
				     auxiliary bound)
			     0
			     (named-continuation k-var))))
	  ((->value cont)
	   (make-lambda 
	    name (cons* k-var f-var s-var required)
	    optional rest auxiliary declarations
	    body*)
	   fuel store (+ ticks cost:lambda)))))))

(define (rewrite/body body fuel store assignable bound ticks cont)
  (if (sequence? body)
      ;; This loop handles assignments from internal DEFINEs specially.
      ;; Simple internal defines are executed for free.  Complex
      ;; internal defines are prohibited because the rest of the MIT
      ;; Scheme system (e.g. unscan-defines) can't handle assignments
      ;; in our introduced continuations.
      (let loop ((actions (sequence-actions body))
		 (actions* '())
		 (ticks  ticks))
	(if (pair? actions)
	    (let ((action (car actions)))
	      (if (and (assignment? action)
		       (memq (assignment-name action) assignable))
		  (let ((name (assignment-name action))
			(value (assignment-value action)))
		    (if (simple? value)
			(rewrite/expression
			 value
			 fuel store bound 0
			 (lambda (value* fuel store ticks*)
			   fuel store	; ignored?
			   (loop (cdr actions)
				 (cons (make-assignment name value*)
				       actions*)
				 (+ ticks ticks*))))
			(error "Internal define too complex"
			       (unsyntax action))))
		  (make-sequence
		   (append (reverse actions*)
			   (list
			    (rewrite/expression (make-sequence actions)
						fuel store bound ticks
						cont))))))
	    (error "No expression in body" (unsyntax body))))
      
      (rewrite/expression body fuel store bound ticks cont)))

(define (rewrite/the-environment expression fuel store bound ticks cont)
  expression fuel store bound ticks cont
  (error "Sorry, (THE-ENVIRONMENT) no allowed" (unsyntax expression)))


(define (rewrite/combination expression fuel store bound ticks cont)
  (let ((operator (combination-operator expression))
	(operands (combination-operands expression)))
    
    (let ((all (cons operator operands)))

      (let walk  ((exprs all)
		  (fuel fuel)
		  (store store)
		  (ticks ticks)
		  (call-site-expressions '())
		  (inner-expressions '()))
	(cond ((null? exprs)
	       (let walk* ((exprs call-site-expressions)
			   (inner? inner-expressions)
			   (fuel fuel)
			   (store store)
			   (ticks ticks)
			   (call-site-expressions '()))
		 (cond ((null? inner?)
			(let ((e* call-site-expressions))
			  (rewrite/jump (car e*) (cdr e*) fuel store ticks cont)))
		       ((car inner?)
			(rewrite/expression
			 (car exprs)
			 fuel store bound ticks
			 (lambda (value fuel store ticks*)
			   (walk* (cdr exprs)
				  (cdr inner?)
				  fuel store ticks*
				  (cons value call-site-expressions)))))
		       (else
			(walk* (cdr exprs) (cdr inner?)
			       fuel store ticks
			       (cons (car exprs) call-site-expressions))))))
	      ((simple? (car exprs))
	       (walk (cdr exprs)
		     fuel store ticks
		     (cons (car exprs) call-site-expressions)
		     (cons #T inner-expressions)))
	      (else
	       (let ((next (generate-uninterned-symbol "receiver."))
		     (value (generate-uninterned-symbol "value.")))
		 (bind next
		       (make-continuation
			value
			(lambda (fuel store)
			  (walk (cdr exprs)
				fuel store 0
				(cons (make-variable value)
				      call-site-expressions)
				(cons #F inner-expressions))))
		       (rewrite/expression
			(car exprs)
			fuel store bound ticks
			(named-continuation next))))))))))

(define (rewrite/combination expression fuel store bound ticks cont)
  (let ((operator (combination-operator expression))
	(operands (combination-operands expression)))

    (define (default)
      (let ((all (cons operator operands)))
	(rewrite/expressions
	 all fuel store bound ticks
	 (lambda (exprs fuel store ticks)
	   (rewrite/jump (car exprs) (cdr exprs) fuel store ticks cont)))))

    (if (and (lambda? operator)
	     (<= ticks cost:max-expression))
	(lambda-components operator
	  (lambda (name required optional rest auxiliary declarations body)
	    (if (and (null? optional)
		     (false? rest)
		     (equal? (length required) (length operands))
		     ;(null? auxiliary)
		     ;(null? declarations)
		     )
		(rewrite/expressions
		 operands fuel store bound ticks
		 (lambda (operands* fuel store ticks)
		   (make-combination
		    (make-lambda
		     name required optional rest auxiliary declarations
		     (rewrite/body
		      body
		      fuel store auxiliary
		      (append required optional (if rest (list rest) ())
			      auxiliary bound)
		      ticks
		      cont))
		    operands*)))
		(default))))
	(default))))


(define (rewrite/expressions expressions fuel store bound ticks cont*)
  ;; cont*: expressions * fuel * store * ticks -> expression
  (let walk  ((exprs expressions)
	      (fuel fuel)
	      (store store)
	      (ticks ticks)
	      (call-site-expressions '())
	      (inner-expressions '()))
    (cond ((null? exprs)
	   (let walk* ((exprs call-site-expressions)
		       (inner? inner-expressions)
		       (fuel fuel)
		       (store store)
		       (ticks ticks)
		       (call-site-expressions '()))
	     (cond ((null? inner?)
		    (let ((e* call-site-expressions))
		      (cont* e* fuel store ticks)))
		   ((car inner?)
		    (rewrite/expression
		     (car exprs)
		     fuel store bound ticks
		     (lambda (value fuel store ticks*)
		       (walk* (cdr exprs)
			      (cdr inner?)
			      fuel store ticks*
			      (cons value call-site-expressions)))))
		   (else
		    (walk* (cdr exprs) (cdr inner?)
			   fuel store ticks
			   (cons (car exprs) call-site-expressions))))))
	  ((simple? (car exprs))
	   (walk (cdr exprs)
		 fuel store ticks
		 (cons (car exprs) call-site-expressions)
		 (cons #T inner-expressions)))
	  (else
	   (let ((next (generate-uninterned-symbol "receiver."))
		 (value (generate-uninterned-symbol "value.")))
	     (bind next
		   (make-continuation
		    value
		    (lambda (fuel store)
		      (walk (cdr exprs)
			    fuel store 0
			    (cons (make-variable value)
				  call-site-expressions)
			    (cons #F inner-expressions))))
		   (rewrite/expression
		    (car exprs)
		    fuel store bound ticks
		    (named-continuation next))))))))


(define (rewrite/jump operator operands fuel store ticks cont)

  (define (make-invocation cont*)
    (let ((cont-expression (make-variable cont*))
	  (fuel-expression
	   (fuel-adjustment fuel (+ ticks (cost:call (length operands))))))
      (let ((stepper
	     (case (length operands)
	       ((0)  (make-variable 'call/0))
	       ((1)  (make-variable 'call/1))
	       ((2)  (make-variable 'call/2))
	       ((3)  (make-variable 'call/3))
	       (else #F))))
	(if stepper
	    (make-combination stepper
			      (cons* cont-expression
				     fuel-expression
				     (make-variable store)
				     (append operands (list operator))))
	    (make-combination (make-variable 'call/n)
			      (cons* operator
				     cont-expression
				     fuel-expression
				     (make-variable store)
				     operands))))))

    (cond ((primitive-procedure? operator)
	   ((->value cont) (make-combination operator operands) fuel store
			   (+ ticks (cost:primitive operator))))
	  ((named-continuation? cont)
	   (make-invocation cont))
	  (else
	   (let ((cont* (generate-uninterned-symbol "use."))
		 (result (generate-uninterned-symbol "result.")))
	     (bind cont*
		   (make-continuation result
		      (lambda (fuel store)
                         (cont (make-variable result) fuel store 0)))
		   (make-invocation cont*))))))			      

(define (simple? expression)
  (cond ((variable? expression) 1)
	((lambda? expression)   1)
	((constant? expression)   1)
	((primitive-procedure? expression) 1)
	((and (combination? expression)
	      (primitive-procedure? (combination-operator expression)))
	 (let loop ((operands (combination-operands expression))
		    (cost     1))
	   (cond ((null? operands) cost)
		 ((simple? (car operands))
		  => (lambda (cost*)
		       (let ((cost (+ cost cost*)))
			 (and (< cost cost:max-expression)
			      (loop (cdr operands) cost)))))
		 (else #F))))
	;;((and (disjunction? expression)
	;;      (simple? (disjunction-predicate expression)))
	;; => (lambda (cost*)
	;;      (cond ((simple? (disjunction-alternative expression))
	;;	     => (lambda (cost**)
	;;		  (< (+ cost* cost**) cost:max-expression)))
	;;	    (else #F))))
	(else #F)))

(define (constant? expression)
  (or (number? expression)
      (string? expression)
      (symbol? expression)
      (vector? expression)
      (eq? expression #T)
      (eq? expression #F)
      (eq? expression '())
      (eq? expression unspecific)
      (char? expression)
      (pair? expression)))

(define (rewrite/comment expression fuel store bound ticks cont)
  (make-comment (comment-text expression)
		(rewrite/expression (comment-expression expression)
				    fuel store bound ticks cont)))


(define (bind name expr body)
  (make-combination
   (make-lambda lambda-tag:let
		(list name) '() #F '() '()
		body)
   (list expr)))


(define (make-continuation var make-body)
  (let ((fuel (make-fuel-var))
	(store (make-store-var)))
    (make-lambda '.cont
		 (list var fuel store) '() #F '() '()
		 (make-body fuel store))))
  

(define (predicate-continuation true-code false-code bound cont)
  (lambda (predicate fuel store ticks)
    (make-conditional   
     predicate
     (rewrite/expression true-code  fuel store bound (+ ticks cost:if) cont)
     (rewrite/expression false-code fuel store bound (+ ticks cost:if) cont))))

(define (rewrite/conditional expression fuel store bound ticks cont)

  (let ((test-cont
	 (predicate-continuation (conditional-consequent expression)
				 (conditional-alternative expression)
				 bound
				 cont)))
    (let ((predicate (conditional-predicate expression)))
      (let ((simple-cost (simple? predicate)))
	(if (and simple-cost
		 (<= (+ simple-cost ticks) cost:max-expression))
	    (rewrite/expression predicate fuel store bound ticks test-cont)
	     (let ((flag (generate-uninterned-symbol "pred."))
		   (test (generate-uninterned-symbol "if.")))
	       (bind test
		     (make-continuation flag
		      (lambda (fuel store)
		        (test-cont (make-variable flag) fuel store 0)))
		     (rewrite/expression
		      predicate
		      fuel store bound ticks
		      (named-continuation test)))))))))


(define (rewrite/delay expression fuel store bound ticks cont)
  expression fuel store bound ticks cont 
  (error "Sorry, DELAY not allowed" (unsyntax expression)))


(define (rewrite/disjunction expression fuel store bound ticks cont)
  (let ((flag (generate-uninterned-symbol "pred."))
	(test (generate-uninterned-symbol "or.")))
    (bind test
	  (make-continuation
	   flag
	   (lambda (fuel store)
	     ((predicate-continuation (make-variable flag)
				      (disjunction-alternative expression)
				      (cons flag bound) ; kludge
				      cont)
	      (make-variable flag)
	      fuel store 0)))
	  (rewrite/expression
	   (disjunction-predicate expression)
	   fuel store bound ticks
	   (named-continuation test)))))


(define (rewrite/in-package expression fuel store bound ticks cont)
  expression fuel store bound ticks cont 
  (error "Sorry, IN-PACKAGE not allowed" (unsyntax expression)))

(define (rewrite/sequence expression fuel store bound ticks cont)
  (let ((actions (sequence-actions expression)))
    (let walk ((next (car actions))
	       (rest (cdr actions))
	       (fuel fuel)
	       (store store)
	       (ticks ticks))
      (cond ((null? rest)
	     (rewrite/expression next fuel store bound ticks cont))
	    ((or (and (simple? next) (< ticks cost:max-sequence))
		 (sharable-expression? next))
	     (rewrite/expression
	      next fuel store bound ticks
	      (lambda (value fuel store ticks)
		(make-sequence (list value
				     (walk (car rest) (cdr rest)
					   fuel store ticks))))))
	    (else
	     (let ((next-name (generate-uninterned-symbol "next."))
		   (value     (generate-uninterned-symbol "ignored.")))
	       (bind next-name
		     (make-continuation
		      value
		      (lambda (fuel store)
			(walk (car rest) (cdr rest) fuel store 0)))
		     (rewrite/expression next fuel store bound ticks
					 (named-continuation next-name)))))))))

(define (rewrite/unassigned? expression fuel store bound ticks cont)
  bound					; ignored
  ((->value cont) expression fuel store (+ ticks cost:unassigned?)))

(define (rewrite/constant expression fuel store bound ticks cont)
  bound					; ignored
  ((->value cont) expression fuel store ticks))


(define (sharable-expression? expression)
  (or (and (definition? expression)
	   (lambda? (definition-value expression)))
      #F))

(define (rewrite/definition expression fuel store bound ticks cont)
  (let ((name  (definition-name expression))
	(value (definition-value expression)))
    (set! *definition-names* (cons name *definition-names*))
    (if (sharable-expression? expression)
	(rewrite/expression
	 value
	 fuel store bound ticks
	 (lambda (value fuel store ticks)
	   (set! *shared-definitions*
		 (cons (make-definition name value) *shared-definitions*))
	   (rewrite/global-definition name (make-variable name)
				      fuel store ticks cont)))
	(rewrite/expression
	 value
	 fuel store bound ticks
	 (lambda (value fuel store ticks)
	   (rewrite/global-definition name value fuel store ticks cont))))))

(define (rewrite/file-contents expression)
  (within-rewriter
   (lambda ()
     (let ((cont  (make-cont-var))
	   (fuel  (make-fuel-var))
	   (store (make-store-var)))
       (let ((file-expression
	      (rewrite/expression expression fuel store '() 0
				  (named-continuation cont))))
	 (make-sequence
	  (append (reverse *shared-definitions*)
		  (list
		   (make-combination
		    (make-variable 'current-simulation.notice-file-definitions)
		    (list *definition-names*)))
		  (list
		   (make-lambda
		    lambda-tag:unnamed
		    (list cont fuel store) '() #F '() '()
		    (make-sequence
		     (list #|(make-combination
			       (make-variable 'declare-linkage)
			       (list (hash-table->alist *cache-map*)))
		           |#
			   file-expression)))))))))))

(define (cps file)
  (let ((scm-file (pathname-new-type file "scm"))
	(tmp-file (pathname-new-type file "tmp"))
	(bin-file (pathname-new-type file "bin")))
    (sf scm-file)
    (rename-file bin-file tmp-file)
    (let ((stuff (fasload tmp-file)))
      (let ((transformed (rewrite/file-contents stuff)))
	(fasdump transformed bin-file)))))

(initialize-package!)
