(declare (usual-integrations))

#|

Macros for hl-sim programs.

(SELECT (event-guard code ...)
	(event-guard => procedure)
	...)

Expands to a call to WAIT-MULTIPLE followed by a dispatch.
|#

(define (select-transformer . clauses)

  (define (check-syntax clause)
    (if (and (list? clause)
	     (or (and (= (length clause) 3)
		      (eq? (second clause) '=>)
		      (not (eq? (car clause) 'ELSE)))
		 (= (length clause) 1)
		 (and (>= (length clause) 2)
		      (not (eq? (second clause) '=>)))))
	'ok
	(error "Illegal SELECT clause" clause)))

  (define (else-clause? clause)
    (and (pair? clause)
	 (eq? (car clause) 'ELSE)))

  (for-each check-syntax clauses)

  (let ((guard-clauses (list-transform-negative clauses else-clause?))
	(else-clauses  (list-transform-positive clauses else-clause?)))
    (if (> (length else-clauses) 1)
	(error "Too many ELSE clauses in SELECT form:" `(select ,@clauses)))

    (let ((event-guards (map car guard-clauses))
	  (result-name  (string->uninterned-symbol "RESULT")))
      (let ((case-clauses
	     (map (lambda (tag clause)
		    (if (eq? (second clause) '=>)
			`((,tag) (,(third clause) (CDR ,result-name)))
			`((,tag) ,@(cdr clause))))
		  (make-initialized-list (length clauses) identity-procedure)
		  guard-clauses)))
      
	`(LET ((,result-name (WAIT-MULTIPLE ,@event-guards)))
	   (IF ,result-name
	       (CASE (CAR ,result-name)
		 ,@case-clauses)
	       ,(if (pair? else-clauses)
		    (if (pair? (cdr (first else-clauses)))
			`(BEGIN ,@(cdr (first else-clauses)))
			'UNSPECIFIC)
		    'UNSPECIFIC)))))))

;; (PARALLEL form form ...)

(define (parallel-transformer . forms)
  (let ((n-forms (length forms)))
    (cond ((= n-forms 0)  '#())
	  ((= n-forms 1)  `(vector ,(car forms)))
	  (else
	   (let ((sync-vars
		  (make-initialized-list (- n-forms 1)
		    (lambda (i)
		      (string->uninterned-symbol
		       (string-append "THREAD-" (number->string (+ i 1))))))))
	     `((LAMBDA ,sync-vars
		 (VECTOR ,(car forms)
			 ,@(map (lambda (v)  `(WAIT ,v)) sync-vars)))
	       ,@(map (lambda (e)
			`(MAKE&START-THREAD (LAMBDA () ,e)))
		      (cdr forms))))))))
					 
	       
(syntax-table/define system-global-syntax-table 'SELECT   select-transformer)
(syntax-table/define system-global-syntax-table 'PARALLEL parallel-transformer)
